"""
Configuration object for the entire pipeline.
"""

import yaml
from typing import List


# Items to read directly from the yaml file with their defaults
direct_read = {
    "auto_plotter_directory": None,
    "auto_plotter_registration": None,
    "observational_data_directory": None,
    "matplotlib_stylesheet": "default",
    "description_template": None,
    "custom_css": None,
}


class Script(object):
    """
    Object describing the core properties of a 'script'.
    """

    # Filename of the python script to be ran
    filename: str
    # Caption for displaying in a webpage, etc.
    caption: str
    # Output file name; required to match up the caption and file in
    # postprocessing
    output_file: str
    # Section heading; used to classify this with similar figures
    # in the output.
    section: str
    # Plot title; written above the caption
    title: str
    # Show on webpage; Defaults to True but used to disable webpage plotting
    # in the config file if required.
    show_on_webpage: bool

    def __init__(self, script_dict: dict):
        """
        Takes the dictionary and extracts it to inner variables.
        """

        self.filename = script_dict.get("filename", "")
        self.caption = script_dict.get("caption", "")
        self.output_file = script_dict.get("output_file", "")
        self.section = script_dict.get("section", "")
        self.title = script_dict.get("title", "")
        self.show_on_webpage = script_dict.get("show_on_webpage", True)

        return

    def __str__(self):
        return (
            f"Script: {self.filename}, produces {self.filename} in {self.section} "
            f"described as {self.caption}"
        )

    def __repr__(self):
        return f"Script object describing the {self.filename} script."


class Config(object):
    """
    Configuration object containing the major parameters read from the
    yaml file.
    """

    # Raw config read directly from the file, before processing.
    raw_config: dict
    scripts: List[Script]

    # Set up the object.
    __slots__ = list(direct_read.keys()) + ["scripts", "config_directory", "raw_config"]

    def __init__(self, config_directory: str):
        """
        Set up and load the config to this object.

        Parameters
        ----------

        config_directory: str
            Directory containing the configuration ``config.yml``.
        """

        self.config_directory = config_directory

        self.__read_config()
        self.__extract_to_variables()
        self.__extract_scripts()

        return

    def __str__(self):
        return f"Configuration file object describing {self.config_directory}"

    def __repr__(self):
        return self.__str__()

    def __read_config(self):
        """
        Read the ``config.yml`` and store the basic contents in
        ``self.raw_config``.
        """

        with open(f"{self.config_directory}/config.yml", "r") as handle:
            self.raw_config = yaml.safe_load(handle)

        return

    def __extract_to_variables(self):
        """
        Extracts items from the configuration dictionary to local variables.
        """

        for variable, default in direct_read.items():
            setattr(self, variable, self.raw_config.get(variable, default))

        return

    def __extract_scripts(self):
        """
        Extracts the items in the scripts section to their own
        list with objects describing each script.
        """

        raw_scripts = self.raw_config.get("scripts", [])
        self.scripts = [Script(script_dict=script_dict) for script_dict in raw_scripts]

        return




"""
Functions that aid in the production of the HTML webpages.
"""

from swiftpipeline import __version__ as pipeline_version
from swiftpipeline.config import Config

from swiftsimio import SWIFTDataset

from velociraptor import __version__ as velociraptor_version
from velociraptor.autoplotter.metadata import AutoPlotterMetadata

from jinja2 import Environment, PackageLoader, FileSystemLoader, select_autoescape
from time import strftime
from typing import List

import unyt


def format_number(number):
    """
    Formats a number from float (with or without units) to a latex-like number.
    """

    try:
        units = f"\\; {number.units.latex_repr}"
    except:
        units = ""

    try:
        mantissa, exponent = ("%.3g" % number).split("e+")
        exponent = f" \\times 10^{{{int(exponent)}}}"
    except:
        mantissa = "%.3g" % number
        exponent = ""

    return f"\\({mantissa}{exponent}{units}\\)"


def get_if_present_float(dictionary, value: str, input_unit=None, output_unit=None):
    """
    A replacement for .get() that also formats the number if present.
    
    Assumes data should be a float.
    """

    try:
        value = float(dictionary[value])

        if input_unit is not None:
            value = unyt.unyt_quantity(value, input_unit)

            if output_unit is not None:
                value.convert_to_units(output_unit)

        return format_number(value)
    except KeyError:
        return ""


def get_if_present_int(dictionary, value: str, input_unit=None, output_unit=None):
    """
    A replacement for .get() that also formats the number if present.
    
    Assumes data should be an integer.
    """

    try:
        value = int(dictionary[value])

        if input_unit is not None:
            value = unyt.unyt_quantity(value, input_unit)

            if output_unit is not None:
                value.convert_to_units(output_unit)

        return format_number(value)
    except KeyError:
        return ""


class WebpageCreator(object):
    """
    Creates webpages based on the information that is provided in 
    the plots metadata through the autoplotter and the additional
    plotting interface provided through the pipeline.
    """

    environment: Environment
    loader: PackageLoader

    variables: dict
    html: str

    auto_plotter_metadata: AutoPlotterMetadata
    config: Config

    def __init__(self):
        """
        Sets up the ``jinja`` templating system.
        """

        self.loader = PackageLoader("swiftpipeline", "templates")
        self.environment = Environment(
            loader=self.loader, autoescape=select_autoescape(["js"])
        )

        # Initialise empty variables dictionary, with the versions of
        # this package and the velociraptor package used.
        self.variables = dict(
            pipeline_version=pipeline_version,
            velociraptor_version=velociraptor_version,
            creation_date=strftime(r"%Y-%m-%d"),
            sections=[],
            runs=[],
        )

        return

    def render_webpage(self, template: str = "plot_viewer.html") -> str:
        """
        Renders a webpage based on the internal variables stored in
        the ``variables`` dictionary.

        Parameters
        ----------

        template: str
            The name of the template that you wish to use. Defaults to
            "plot_viewer.html".

        Returns
        -------

        html: str
            The resulting HTML. This is also stored in ``.html``.
        """

        self.html = self.environment.get_template(template, parent="base.html").render(
            **self.variables
        )

        return self.html

    def add_metadata(self, page_name: str):
        """
        Add additional metadata to the page.

        Parameters
        ----------

        page_name: str
            Name to put in the page title.
        """

        self.variables.update(dict(page_name=page_name))

    def add_auto_plotter_metadata(self, auto_plotter_metadata: AutoPlotterMetadata):
        """
        Adds the auto plotter metadata to the section / plot metadata.

        Parameters
        ----------

        auto_plotter_metadata: AutoPlotterMetadata
            The complete instance of ``AutoPlotterMetadata`` after running the
            ``velociraptor`` ``AutoPlotter``. 
        """

        self.auto_plotter_metadata = auto_plotter_metadata

        # Unique sections
        sections = {plot.section for plot in auto_plotter_metadata.metadata}

        for section in sections:
            plots = [
                dict(
                    filename=f"{plot.filename}.{auto_plotter_metadata.file_extension}",
                    title=plot.title,
                    caption=plot.caption,
                )
                for plot in auto_plotter_metadata.metadata
                if plot.section == section and plot.show_on_webpage
            ]

            self.variables["sections"].append(
                dict(title=section, plots=plots, id=abs(hash(section)))
            )

        return

    def add_config_metadata(self, config: Config):
        """
        Adds the section metadata from the additional plots
        defined under ``config.yml::scripts``.

        Parameters
        ----------

        config: Config
            Configuration object from ``swift-pipeline``.
        """

        self.config = config

        # Unique sections
        sections = {script.section for script in config.scripts}

        for section in sections:
            plots = [
                dict(
                    filename=script.output_file,
                    title=script.title,
                    caption=script.caption,
                )
                for script in config.scripts
                if script.section == section and script.show_on_webpage
            ]

            self.variables["sections"].append(
                dict(title=section, plots=plots, id=abs(hash(section)))
            )

        return

    def add_run_metadata(self, config: Config, snapshots: List[SWIFTDataset]):
        """
        Adds the "run" metadata (using the user-defined description.html).

        Parameters
        ----------

        config: Config
            Configuration object from ``swift-pipeline``.

        snapshots: List[SWIFTDataset].
            SWIFT Datasets used to generate the HTML with.
        """

        loader = FileSystemLoader(config.config_directory)
        environment = Environment(loader=loader)

        environment.filters["format_number"] = format_number
        environment.filters["get_if_present_float"] = get_if_present_float
        environment.filters["get_if_present_int"] = get_if_present_int

        if config.description_template is not None:
            self.variables["runs"] = [
                dict(
                    description=environment.get_template(
                        config.description_template
                    ).render(data=data),
                )
                for data in snapshots
            ]

        if config.custom_css is not None:
            self.variables["custom_css"] = environment.get_template(
                config.custom_css
            ).render()

        return

    def save_html(self, filename: str):
        """
        Saves the html in ``self.html`` to the filename provided.

        Parameters
        ----------

        filename: str
            Full filename (including file path) to save the HTML as.
        """

        with open(filename, "w") as handle:
            handle.write(self.html)
